﻿using System;
using System.ComponentModel;
using System.Globalization;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Threading;

namespace BandedSpectrumAnalyzer
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        #region Fields
        private DispatcherTimer timer = new DispatcherTimer();
        private bool inTimerPositionUpdate;
        #endregion

        #region Dependency Properties
        public static readonly DependencyProperty ScaleValueProperty;

        public float ScaleValue
        {
            get { return (float)GetValue(ScaleValueProperty); }
            set { SetValue(ScaleValueProperty, value); }
        }
        #endregion

        #region Dependency Property Validators
        private static bool IsScaleValid(object value)
        {
            float val = (float)value;
            return (val >= 0);
        }

        private static object CoerceScale(DependencyObject d, object value)
        {
            float val = (float)value;

            if (float.IsNaN(val))
                return 1.0f;

            val = Math.Max(val, 0.5f);
            val = Math.Min(val, 5.0f);
            return val;
        }
        #endregion

        #region Constructors
        static MainWindow()
        {
            FrameworkPropertyMetadata metadata = new FrameworkPropertyMetadata(1.0f, FrameworkPropertyMetadataOptions.AffectsArrange);
            metadata.CoerceValueCallback = new CoerceValueCallback(MainWindow.CoerceScale);
            ScaleValueProperty = DependencyProperty.Register("ScaleValue", typeof(float), typeof(MainWindow), metadata, new ValidateValueCallback(MainWindow.IsScaleValid));
        }

        public MainWindow()
        {
            InitializeComponent();

            BassEngine bassEngine = BassEngine.Instance;

            bassEngine.PropertyChanged += BassEngine_PropertyChanged;
            UIHelper.Bind(bassEngine, "CanStop", StopButton, Button.IsEnabledProperty);
            UIHelper.Bind(bassEngine, "CanPlay", PlayButton, Button.IsEnabledProperty);
            UIHelper.Bind(bassEngine, "CanPause", PauseButton, Button.IsEnabledProperty);
            UIHelper.Bind(bassEngine, "ChannelLength", ProgressSlider, Slider.MaximumProperty);

            LoadDarkTheme();
            LoadBlueTheme();
        }
        #endregion

        #region Bass Engine Events
        private void BassEngine_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            BassEngine engine = BassEngine.Instance;
            switch (e.PropertyName)
            {
                case "FileTag":
                    if (engine.FileTag != null)
                    {
                        TagLib.Tag tag = engine.FileTag.Tag;
                        AlbumText.Text = tag.Album;
                        ArtistText.Text = tag.AlbumArtists.Length > 0 ? tag.AlbumArtists[0] : string.Empty;
                        TitleText.Text = tag.Title;
                        YearText.Text = tag.Year.ToString(CultureInfo.InvariantCulture);
                        GenreText.Text = tag.Genres.Length > 0 ? tag.Genres[0] : string.Empty;
                        TrackText.Text = tag.Track.ToString(CultureInfo.InvariantCulture);
                        DiscText.Text = tag.Disc.ToString(CultureInfo.InvariantCulture);

                        if (tag.Pictures.Length > 0)
                        {
                            using (MemoryStream albumArtworkMemStream = new MemoryStream(tag.Pictures[0].Data.Data))
                            {
                                BitmapImage albumImage = new BitmapImage();
                                albumImage.BeginInit();
                                albumImage.CacheOption = BitmapCacheOption.OnLoad;
                                albumImage.StreamSource = albumArtworkMemStream;
                                albumImage.EndInit();
                                AlbumArtPanel.AlbumArtImage = albumImage;
                                albumArtworkMemStream.Close();
                            }
                        }
                        else
                        {
                            AlbumArtPanel.AlbumArtImage = null;
                        }
                    }
                    else
                    {
                        AlbumText.Text = string.Empty;
                        ArtistText.Text = string.Empty;
                        TitleText.Text = string.Empty;
                        YearText.Text = string.Empty;
                        GenreText.Text = string.Empty;
                        TrackText.Text = string.Empty;
                        DiscText.Text = string.Empty;
                        AlbumArtPanel.AlbumArtImage = null;
                    }
                    break;
                case "ChannelPosition":
                    inTimerPositionUpdate = true;
                    ProgressSlider.Value = engine.ChannelPosition;
                    inTimerPositionUpdate = false;
                    break;
            }

        }
        #endregion

        #region UI Events
        private void PlayButton_Click(object sender, RoutedEventArgs e)
        {
            if (BassEngine.Instance.CanPlay)
                BassEngine.Instance.Play();
        }

        private void PauseButton_Click(object sender, RoutedEventArgs e)
        {
            if (BassEngine.Instance.CanPause)
                BassEngine.Instance.Pause();
        }

        private void StopButton_Click(object sender, RoutedEventArgs e)
        {
            if (BassEngine.Instance.CanStop)
                BassEngine.Instance.Stop();
        }

        private void BlackBackgroundButton_Click(object sender, RoutedEventArgs e)
        {
            LoadDarkTheme();
        }

        private void WhiteBackgroundButton_Click(object sender, RoutedEventArgs e)
        {
            LoadLightTheme();
        }

        private void GreenThemeButton_Click(object sender, RoutedEventArgs e)
        {
            LoadGreenTheme();
        }

        private void BlueThemeButton_Click(object sender, RoutedEventArgs e)
        {
            LoadBlueTheme();
        }

        private void OrangeThemeButton_Click(object sender, RoutedEventArgs e)
        {
            LoadOrangeTheme();
        }

        private void RedThemeButton_Click(object sender, RoutedEventArgs e)
        {
            LoadRedTheme();
        }

        private void BrowseButton_Click(object sender, RoutedEventArgs e)
        {
            Microsoft.Win32.OpenFileDialog openDialog = new Microsoft.Win32.OpenFileDialog();
            openDialog.Filter = "(*.mp3, *.m4a)|*.mp3;*.m4a";
            if (openDialog.ShowDialog() == true)
            {
                BassEngine.Instance.OpenFile(openDialog.FileName);
                FileText.Text = openDialog.FileName;
            }
        }

        private void ProgressSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if (!inTimerPositionUpdate)
            {
                BassEngine.Instance.SetChannelPosition((long)e.NewValue);
            }
        }
        #endregion

        #region Themes
        private void LoadDarkTheme()
        {
            this.Resources.MergedDictionaries.Clear();
            ResourceDictionary themeResources = Application.LoadComponent(new Uri("DarkTheme.xaml", UriKind.Relative)) as ResourceDictionary;
            this.Resources.MergedDictionaries.Add(themeResources);
            this.Background = FindResource("GradientWindowBackgroundBrush") as LinearGradientBrush;
            SetValueTextColor(Colors.White);
        }

        private void LoadLightTheme()
        {
            this.Resources.MergedDictionaries.Clear();
            ResourceDictionary themeResources = Application.LoadComponent(new Uri("LightTheme.xaml", UriKind.Relative)) as ResourceDictionary;
            this.Resources.MergedDictionaries.Add(themeResources);
            this.Background = FindResource("GradientWindowBackgroundBrush") as LinearGradientBrush;
            SetValueTextColor(Colors.Black);
        }

        private void LoadGreenTheme()
        {
            spectrumAnalyzer.PeakBrush = new SolidColorBrush(Color.FromRgb(137, 236, 106));
            spectrumAnalyzer.BarBrush = new LinearGradientBrush(Color.FromRgb(34, 141, 0), Color.FromRgb(52, 216, 0), new Point(0, 1), new Point(0, 0));
            SetHeaderTextColor(Color.FromRgb(52, 216, 0));
        }

        private void LoadBlueTheme()
        {
            spectrumAnalyzer.PeakBrush = new SolidColorBrush(Color.FromRgb(97, 180, 207));
            spectrumAnalyzer.BarBrush = new LinearGradientBrush(Color.FromRgb(2, 78, 104), Color.FromRgb(6, 121, 159), new Point(0, 1), new Point(0, 0));
            SetHeaderTextColor(Color.FromRgb(6, 121, 159));
        }

        private void LoadOrangeTheme()
        {
            spectrumAnalyzer.PeakBrush = new SolidColorBrush(Color.FromRgb(255, 186, 115));
            spectrumAnalyzer.BarBrush = new LinearGradientBrush(Color.FromRgb(166, 84, 0), Color.FromRgb(255, 129, 0), new Point(0, 1), new Point(0, 0));
            SetHeaderTextColor(Color.FromRgb(255, 129, 0));
        }

        private void LoadRedTheme()
        {
            spectrumAnalyzer.PeakBrush = new SolidColorBrush(Color.FromRgb(255, 125, 115));
            spectrumAnalyzer.BarBrush = new LinearGradientBrush(Color.FromRgb(166, 12, 0), Color.FromRgb(255, 19, 0), new Point(0, 1), new Point(0, 0));
            SetHeaderTextColor(Color.FromRgb(255, 19, 0));
        }

        private void SetHeaderTextColor(Color color)
        {
            ArtistLabel.Foreground = new SolidColorBrush(color);
            AlbumLabel.Foreground = new SolidColorBrush(color);
            TitleLabel.Foreground = new SolidColorBrush(color);
            YearLabel.Foreground = new SolidColorBrush(color);
            GenreLabel.Foreground = new SolidColorBrush(color);
            TrackLabel.Foreground = new SolidColorBrush(color);
            DiscLabel.Foreground = new SolidColorBrush(color);
        }

        private void SetValueTextColor(Color color)
        {
            ArtistText.Foreground = new SolidColorBrush(color);
            AlbumText.Foreground = new SolidColorBrush(color);
            TitleText.Foreground = new SolidColorBrush(color);
            YearText.Foreground = new SolidColorBrush(color);
            GenreText.Foreground = new SolidColorBrush(color);
            TrackText.Foreground = new SolidColorBrush(color);
            DiscText.Foreground = new SolidColorBrush(color);
            BackgroundLabel.Foreground = new SolidColorBrush(color);
            BarColorLabel.Foreground = new SolidColorBrush(color);
        }
        #endregion
    }
}
